@@ -6,11 +6,16 @@ import os
 import asyncio
 import tempfile
 
+from io import BytesIO
+from collections.abc import AsyncGenerator
+
 import numpy as np
 from PIL import Image
-from skimage import draw
 
-from collections.abc import AsyncGenerator
+import scipy.ndimage
+from skimage import draw, morphology 
+
+from tiatoolbox.tools import tissuemask 
 
 from omero.gateway import _BlitzGateway, ImageWrapper, FileAnnotationWrapper, ShapeWrapper
 import omero.model.enums as omero_enums
@@ -24,7 +29,7 @@ from omero_model_RectangleI import RectangleI
 from omero_model_FileAnnotationI import FileAnnotationI
 
 from lavlab import omero_asyncio
-from lavlab.python_util import chunkify, merge_async_iters, interlace_lists, lookup_filetype_by_name, FILETYPE_DICTIONARY, rgba_to_uint, uint_to_rgba
+from lavlab.python_util import chunkify, merge_async_iters, interlace_lists, lookup_filetype_by_name, FILETYPE_DICTIONARY, rgba_to_uint, uint_to_rgba, create_array
 
 PARALLEL_STORE_COUNT=4
 """Number of pixel stores to be created for an image."""
@@ -134,7 +139,8 @@ asyncio.run(work(img, tiles, res_lvl, dims))
         # set res and get default res level if necessary
         if resLvl is None:
             resLvl = await rps.getResolutionLevels()
-        await rps.setResolutionLevel(resLvl-1)
+            resLvl -= 1
+        await rps.setResolutionLevel(resLvl)
 
         # request and return tiles
         i=1
@@ -149,10 +155,10 @@ asyncio.run(work(img, tiles, res_lvl, dims))
 
     # force async client
     session =  omero_asyncio.AsyncSession(img._conn.c.sf)
-
+    session.setSecurityContext(img.details.group)
     # create parallel raw pixels stores
     jobs=[]
-    for chunk in chunkify(tiles, int(len(tiles)/PARALLEL_STORE_COUNT)):
+    for chunk in chunkify(tiles, PARALLEL_STORE_COUNT):
         jobs.append(work(img.getPrimaryPixels().getId(), chunk, resLvl))
     return merge_async_iters(*jobs)
 
@@ -260,10 +266,6 @@ def getChannelsAtResolution(img: ImageWrapper, xy_dim: tuple[int,int], channels:
     """
 Gathers tiles and scales down to desired resolution.
 
-Warns
--------
-Out of Memory issues ahead! Request a reasonable resolution!
-
 Parameters
 ----------
 img: omero.gateway.ImageWrapper
@@ -284,7 +286,7 @@ PIL.Image.Image
         images = []
         for channel in channels:
             tiles = createTileList2D(0,channel,0,*xy_info)
-            arr = np.zeros((xy_info[1], xy_info[0]), np.uint8)
+            arr = create_array((xy_info[1], xy_info[0]), np.uint8)
             async for tile, (z,c,t,coord) in getTiles(img,tiles,res_lvl):
                 arr [
                     coord[1]:coord[1]+coord[3],
@@ -296,17 +298,17 @@ PIL.Image.Image
                 image = image.resize(xy)
             images.append(image)
         return images
-
-    return asyncio.run(work(img, xy_dim, channels))
+    
+    event_loop = asyncio._get_running_loop()
+    if event_loop is None:
+        return asyncio.run(work(img, xy_dim, channels))
+    else:
+        return work(img,xy_dim,channels)
 
 def getImageAtResolution(img: ImageWrapper, xy_dim: tuple[int,int]) -> Image.Image:
     """
 Gathers tiles of full rgb image and scales down to desired resolution.
 
-Warns
--------
-Out of Memory issues ahead! Request a reasonable resolution!
-
 Parameters
 ----------
 img: omero.gateway.ImageWrapper
@@ -323,7 +325,7 @@ PIL.Image.Image
         res_lvl, xy_info = getClosestResolutionLevel(img, xy)
         tiles = createFullTileList([0,],range(3),[0,], xy_info[0],xy_info[1], xy_info[2:])# rgb tile list
 
-        arr = np.zeros((xy_info[1], xy_info[0], 3), np.uint8)
+        arr = create_array((xy_info[1], xy_info[0], 3), np.uint8)
         async for tile, (z,c,t,coord) in getTiles(img,tiles,res_lvl):
             arr [
                 coord[1]:coord[1]+coord[3],
@@ -337,8 +339,11 @@ PIL.Image.Image
 
         return image
 
-    return asyncio.run(work(img, xy_dim))
-
+    event_loop = asyncio._get_running_loop()
+    if event_loop is None:
+        return asyncio.run(work(img, xy_dim))
+    else:
+        return work(img, xy_dim)
 
 def getLargeRecon(img:ImageWrapper, downsample_factor:int = 10, workdir='./', skip_upload=False):
     """
@@ -390,6 +395,88 @@ str
 
     return recon, recon_img
 
+
+def loadFullImageSmart(img_obj: ImageWrapper):
+    """
+Attempts to only request tiles with tissue, with the rest being filled in by white space.
+    """
+    
+    async def work(img, mask):
+        # Overall image dimensions
+        image_width, image_height = img_obj.getSizeX(), img_obj.getSizeY()
+
+        # Scaling factors
+        scale_x = mask.shape[1] / image_width
+        scale_y = mask.shape[0] / image_height
+
+        tiles = createTileListFromImage(img)
+        arr = create_array((image_height, image_width, img.getSizeC()), np.uint8)
+        # Empty list to store tiles that land on the mask
+        tiles_on_land = []
+
+        for z,c,t,tile in tiles:
+            x, y, width, height = tile
+
+            # Calculate downscaled coordinates and dimensions
+            x_ds, y_ds = int(x * scale_x), int(y * scale_y)
+            width_ds, height_ds = int(width * scale_x), int(height * scale_y)
+
+            # Check if any pixel in the corresponding mask area is True (assuming binary mask)
+            if np.any(mask[y_ds:(y_ds+height_ds), x_ds:(x_ds+width_ds)]):
+                tiles_on_land.append((z,c,t,tile))
+        async for tile, (z,c,t,coord) in getTiles(img,tiles_on_land):
+            arr [
+                coord[1]:coord[1]+coord[3],
+                coord[0]:coord[0]+coord[2],
+                c ] = tile
+        return Image.fromarray(arr)
+    
+    mask = maskTissueLoose(img_obj)
+
+    event_loop = asyncio._get_running_loop()
+    if event_loop is None:
+        return asyncio.run(work(img_obj, mask))
+    else:
+        return work(img_obj, mask)
+
+
+#
+## MASKING
+#
+
+def maskTissueLoose(img_obj:ImageWrapper, mpp=728):
+    phys_w = img_obj.getPixelSizeX()
+    downsample_factor = mpp / phys_w
+    scaled_dims = getDownsampledXYDimensions(img_obj, downsample_factor)
+
+    # get img ( at super low res )
+    img = Image.open(BytesIO(img_obj.getThumbnail(scaled_dims)))
+    arr = np.array(img)
+    
+    # # tia tissue masker (too fine for our purposes)
+    mask = tissuemask.MorphologicalMasker(mpp=mpp).fit_transform([arr])[0]
+
+    # clean up mask
+    mask = morphology.remove_small_holes(mask)
+    mask = morphology.remove_small_objects(mask)
+
+    # increase resolution
+    scale = 32/mpp 
+    mask_img=Image.fromarray(mask)
+    full_mask_img = mask_img.resize((int(mask_img.size[0]/scale), int(mask_img.size[1]/scale)))
+    mask_img.close()
+    mask = np.array(full_mask_img)
+    full_mask_img.close()
+    
+    # smooth up mask
+    mask = scipy.ndimage.binary_dilation(mask, iterations=16)
+    mask = scipy.ndimage.gaussian_filter(mask.astype(float), sigma=24)
+    mask = mask > 0.5
+
+    # invert mask
+    return ~mask
+
+
 #
 ## TILES
 #
@@ -540,6 +627,31 @@ list
 #
 ## ROIS
 #
+
+def getRois(img: ImageWrapper, roi_service=None):
+    """
+Gathers OMERO RoiI objects.
+
+Parameters
+----------
+img: omero.gateway.ImageWrapper
+    Omero Image object from conn.getObjects()
+roi_service: omero.RoiService, optional
+    Allows roiservice passthrough for performance
+    """
+    if roi_service is None:
+        roi_service = img._conn.getRoiService()
+        close_roi = True
+    else:
+        close_roi = False
+
+    rois = roi_service.findByImage(img.getId(), None, img._conn.SERVICE_OPTS).rois
+
+    if close_roi:
+        roi_service.close()
+
+    return rois
+
 def getShapesAsPoints(img: ImageWrapper, point_downsample=4, img_downsample=1,
                       roi_service=None) -> list[tuple[int, tuple[int,int,int], list[tuple[float, float]]]]:
     """
@@ -561,18 +673,13 @@ Returns
 returns: list[ shape.id, (r,g,b), list[tuple(x,y)] ]
     list of tuples containing a shape's id, rgb value, and a tuple of row and column points
     """
-    if roi_service is None:
-        roi_service=img._conn.getRoiService()
-        close_roi=True
 
     sizeX = img.getSizeX() / img_downsample
     sizeY = img.getSizeY() / img_downsample
     yx_shape = (sizeY,sizeX)
 
-    result = roi_service.findByImage(img.getId(), None)
-
     shapes=[]
-    for roi in result.rois:
+    for roi in getRois(img, roi_service):
         points= None
         for shape in roi.copyShapes():
 
@@ -613,8 +720,6 @@ returns: list[ shape.id, (r,g,b), list[tuple(x,y)] ]
     if not shapes : # if no shapes in shapes return none
         return None
 
-    if close_roi: roi_service.close()
-
     # make sure is in correct order
     return sorted(shapes)
 
